"""
Some rudimentary tests. Instantiates a Plot object, but doesn't test plotting.

Discovered that this set of rudimentary tests might be necessary after seeing
the following issue:

    https://github.com/ericmjl/nxviz/issues/160
"""

from .test_utils import diff_plots, corresponding_lists

from random import random

import networkx as nx

from nxviz import ArcPlot, CircosPlot, GeoPlot, MatrixPlot

from matplotlib.testing.decorators import _image_directories

G = nx.erdos_renyi_graph(n=20, p=0.2)

G_geo = G.copy()
for n, d in G_geo.nodes(data=True):
    G_geo.nodes[n]["latitude"] = random()
    G_geo.nodes[n]["longitude"] = random()
    G_geo.nodes[n]["dpcapacity"] = random()

baseline_dir, result_dir = _image_directories(lambda: "dummy func")


def test_circos_plot():
    c = CircosPlot(G)  # noqa: F841
    diff = diff_plots(c, "circos.png", baseline_dir, result_dir)
    assert diff is None


def test_matrix_plot():
    m = MatrixPlot(G)  # noqa: F841
    diff = diff_plots(m, "matrix.png", baseline_dir, result_dir)
    assert diff is None


def test_arc_plot():
    a = ArcPlot(G)  # noqa: F841
    diff = diff_plots(a, "arc.png", baseline_dir, result_dir)
    assert diff is None


def test_geo_plot():
    g = GeoPlot(
        G_geo,
        node_lat="latitude",
        node_lon="longitude",  # noqa: F841
        color="dpcapacity",
    )
    diff = diff_plots(g, "geo.png", baseline_dir, result_dir)
    assert diff is None


def test_plot_size():
    c = CircosPlot(G, figsize=(3, 3))  # noqa: F841
    diff = diff_plots(c, "circos33.png", baseline_dir, result_dir)
    assert diff is None


def test_edge_widths():
    # add weight as attribute and fill with random numbers
    edges = G.edges()
    for u, v in edges:
        G[u][v]["weight"] = random()
    # also extract list for testing
    weights = [G[u][v]["weight"] for u, v in edges]
    # add weights as property
    c = CircosPlot(G, edge_width="weight")
    assert c.edge_widths == weights
    a = ArcPlot(G, edge_width="weight")
    assert a.edge_widths == weights
    # add weights as list
    c = CircosPlot(G, edge_width=weights)
    assert c.edge_widths == weights
    a = ArcPlot(G, edge_width=weights)
    assert a.edge_widths == weights


def test_edge_color():
    # add color as attribute and fill with random numbers
    edges = G.edges()
    for u, v in edges:
        G[u][v]["type"] = "a" if random() < 0.5 else "b"
    # also extract list for testing
    types = [G[u][v]["type"] for u, v in edges]
    # add color as property
    c = CircosPlot(G, edge_color="type")
    assert corresponding_lists(c.edge_colors, types)
    a = ArcPlot(G, edge_color="type")
    assert corresponding_lists(a.edge_colors, types)


def test_node_size():
    # add size as attribute and fill with random numbers
    nodes = G.nodes()
    for u in nodes:
        G.nodes[u]["score"] = random()
    # also extract list for testing
    scores = [G.nodes[u]["score"] for u in nodes]
    # add color as property
    a = ArcPlot(G, node_size="score")
    assert a.node_sizes == scores
    # add types as list
    a = ArcPlot(G, node_size=scores)
    assert a.node_sizes == scores



"""Tests for geometry module."""

import numpy as np
from random import choice

import nxviz.polcart as polcart
from hypothesis import assume, given, settings
from hypothesis.strategies import floats, integers, lists
from nxviz.geometry import (
    circos_radius,
    correct_negative_angle,
    get_cartesian,
    node_theta,
)


def test_circos_radius():
    """
    Check radius correctness.

    Uses the other triangle geometry rule to check that the radius is correct.
    """
    n_nodes = 10
    node_r = 1

    A = 2 * np.pi / n_nodes  # noqa

    circ_r = 2 * node_r / np.sqrt(2 * (1 - np.cos(A)))

    assert np.allclose(circ_r, circos_radius(n_nodes, node_r))


# @settings(perform_health_check=False)
@given(lists(integers()))
def test_node_theta(nodelist):
    """Tests node_theta function."""
    assume(len(nodelist) > 0)
    node = choice(nodelist)
    theta_obs = node_theta(nodelist, node)

    i = nodelist.index(node)
    theta_exp = -np.pi + i * 2 * np.pi / len(nodelist)
    if theta_exp > np.pi:
        theta_exp = np.pi - theta_exp
    assert np.allclose(theta_obs, theta_exp)


@given(floats(), floats())
def test_get_cartesian(r, theta):
    """
    Test for get_cartesian.

    Makes sure that `get_cartesian` remains a wrapper around polcart's
    `to_cartesian`.
    """
    assume(np.isfinite(theta))
    assume(np.isfinite(r))
    assert get_cartesian(r, theta) == polcart.to_cartesian(r, theta)


# @settings(perform_health_check=False)
@given(floats())
def test_correct_negative_angle(angle):
    """Test for correct calculation of negative angle."""
    assume(angle < 0)
    assume(angle >= -2 * np.pi)
    exp = 2 * np.pi + angle
    obs = correct_negative_angle(angle)

    assert np.allclose(obs, exp)
    assert obs <= 2 * np.pi
    assert obs >= 0



import pandas as pd
from networkx import Graph, MultiGraph


def graph_from_dataframe(
    dataframe,
    threshold_by_percent_unique=0.1,
    threshold_by_count_unique=None,
    node_id_columns=[],
    node_property_columns=[],
    edge_property_columns=[],
    node_type_key="type",
    edge_type_key="type",
    collapse_edges=True,
    edge_agg_key="weight",
):
    """
    Build an undirected graph from a pandas dataframe.

    This function attempts to infer which cells should become nodes
    based on either:

        a. what percentage of the column are unique values (defaults to 10%)
        b. an explicit count of unique values (i.e. any column with 7 unique
           values or less)
        c. an explicit list of column keys (i.e.
           ['employee_id', 'location_code'])

    Column headers are preserved as node and edge 'types'. By default, this is
    stored using the key 'type' which is used by some graph import processes
    but can be reconfigured.

    This function uses a MultiGraph structure during the build phase so that it
    is possible to make multiple connections between nodes. By default, at the
    end of the build phase, the MultiGraph is converted to a Graph and the
    count of edges between each node-pair is written as a 'weight' property.

    :param pandas.DataFrame dataframe: A pandas dataframe containing the data
        to be converted into a graph.
    :param float threshold_by_percent_unique: A percent value used to determine
        whether a column should be used to generate nodes based on its
        cardinality (i.e. in a dataframe with 100 rows, treat any column with
        10 or less unique values as a node)
    :param int threshold_by_count_unique: A numeric value used to determine
        whether a column should be used to generate nodes based on its
        cardinality (i.e. if 7 is supplied, treat any column with 7 or less
        unique values as a node) - supplying a value will take priority over
        percent_unique
    :param list node_id_columns: A list of column headers to use for generating
        nodes. Suppyling any value will take precedence over
        threshold_by_percent_unique or threshold_by_count_unique. Note: this
        can cause the size of the graph to expand significantly since every
        unique value in a column will become a node.
    :param list node_property_columns: A list of column headers to use for
        generating properties of nodes. These can include the same column
        headers used for the node id.
    :param list edge_property_columns: A list of column headers to use for
        generating properties of edges.
    :param str node_type_key: A string that sets the key will be used to
        preserve the column name as node property (this is useful for importing
        networkx graphs to databases that distinguish between node 'types' or
        for visually encoding those types in plots).
    :param str edge_type_key: A string that sets the key will be used to keep
        track of edge relationships an 'types' (this is useful for importing
        networkx graphs to databases that distinguish between edge'types' or
        for visually encoding those types in plots). Edge type values are
        automatically set to <node_a_id>_<node_b_id>.
    :param bool collapse_edges: Graphs are instantiated as a 'MultiGraph'
        (allow multiple edges between nodes) and then collapsed into a 'Graph'
        which only has a single edge between any two nodes. Information is
        preserved by aggregating the count of those edges as a 'weight' value.
        Set this value to False to return the MultiGraph. Note: this can cause
        the size of the graph to expand significantly since each row can
        potentially have n! edges where n is the number of columns in the
        dataframe.
    :param str edge_agg_key: A string that sets the key the edge count will be
        assigned to when edges are aggregated.
    :returns: A networkx Graph (or MultiGraph if collapse_edges is set to
        False).
    """

    assert isinstance(
        dataframe, pd.DataFrame
    ), "{} is not a pandas DataFrame".format(dataframe)

    M = MultiGraph()

    # if explicit specification of node_id_columns is provided, use those
    if len(node_id_columns) > 0:
        node_columns = node_id_columns
    else:
        # otherwise, compute with thresholds based on the dataframe
        if threshold_by_count_unique:
            node_columns = sorted(
                [
                    col
                    for col in dataframe.columns
                    if dataframe[col].nunique() <= threshold_by_count_unique
                ]
            )
        else:
            node_columns = sorted(
                [
                    col
                    for col in dataframe.columns
                    if dataframe[col].nunique() / dataframe.shape[0]
                    <= threshold_by_percent_unique  # NOQA to preserve meaningful variable names
                ]
            )

    # use the unique values for each node column as node types
    for node_type in node_columns:
        M.add_nodes_from(
            [
                (node, {node_type_key: node_type})
                for node in dataframe[node_type].unique()
            ]
        )

    # iterate over the rows and generate an edge for each pair of node columns
    for i, row in dataframe.iterrows():
        # assemble the edge properties as a dictionary
        edge_properties = {k: row[k] for k in edge_property_columns}

        # iterate over the node_ids in each node_column of the dataframe row
        node_buffer = []
        for node_type in node_columns:
            node_id = row[node_type]

            # get a reference to the node and assign any specified properties
            node = M.nodes[node_id]
            for k in node_property_columns:
                # if values are not identical, append with a pipe delimiter
                if k not in node:
                    node[k] = row[k]
                elif isinstance(node[k], str) and str(row[k]) not in node[k]:
                    node[k] += "|{}".format(str(row[k]))
                elif str(row[k]) not in str(node[k]):
                    node[k] = str(node[k]) + "|{}".format(str(row[k]))

            # build edges using precomputed edge properties
            for other_node_id, other_node_type in node_buffer:
                # sort node_type so undirected edges all share the same type
                ordered_name = "_".join(sorted([node_type, other_node_type]))
                edge_properties[edge_type_key] = ordered_name
                M.add_edge(node_id, other_node_id, **edge_properties)

            # store the node from this column in the buffer for future edges
            node_buffer.append((node_id, node_type))

    if collapse_edges:
        # convert the MultiGraph to a Graph
        G = Graph(M)
        k = edge_agg_key
        # preserve the edge count as a sum of the weight values
        for u, v, data in M.edges(data=True):
            w = data[k] if k in data else 1.0
            edge = G[u][v]
            edge[k] = (w + edge[k]) if k in edge else w
        return G

    return M



import pytest

import os

import matplotlib.pyplot as plt

from matplotlib.testing.compare import compare_images

from nxviz.utils import (
    infer_data_type,
    is_data_diverging,
    is_data_homogenous,
    is_groupable,
    num_discrete_groups,
)

categorical = ["sun", "moon", "light"]
ordinal = [1, 2, 3, 4, 5]
mixed = ["sun", 1, 2.5]
continuous = [1.0, 1.1, 1.2]
diverging_ordinal = [1, 2, 3, 4, -1, -2, -3, -4]
diverging_continuous = [0.0, 0.1, 0.2, 0.3, -0.1, -0.2, -0.3]
unknown_type = [(1, 2), (2, 3), (3, 4)]
binomial = ["hello", "interesting"]
binomial_integer = [0, 1]
binomial_float = [0.1, 0.2]


def test_is_data_homogenous():
    assert not is_data_homogenous(mixed)
    assert is_data_homogenous(categorical)
    assert is_data_homogenous(ordinal)
    assert is_data_homogenous(continuous)


def test_infer_data_type():
    assert infer_data_type(categorical) == "categorical"
    assert infer_data_type(ordinal) == "ordinal"
    assert infer_data_type(continuous) == "continuous"


def test_is_data_diverging():
    assert is_data_diverging(diverging_ordinal)
    assert is_data_diverging(diverging_continuous)

    assert not is_data_diverging(ordinal)
    assert not is_data_diverging(continuous)


def test_unknown_data_type():
    with pytest.raises(ValueError):
        infer_data_type(unknown_type)


def test_is_groupable():
    assert is_groupable(categorical)
    with pytest.raises(AssertionError):
        is_groupable(mixed)


def test_num_discrete_groups():
    assert num_discrete_groups(categorical) == 3
    assert num_discrete_groups(ordinal) == 5


def test_binomial():
    assert infer_data_type(binomial) == "categorical"
    assert infer_data_type(binomial_float) == "categorical"
    assert infer_data_type(binomial_integer) == "categorical"


def diff_plots(plot, plot_fn, baseline_dir, result_dir):
    plot.draw()
    img_fn = plot_fn
    baseline_img_path = os.path.join(baseline_dir, img_fn)
    result_img_path = os.path.join(result_dir, img_fn)
    plt.savefig(result_img_path)
    diff = compare_images(baseline_img_path, result_img_path, tol=0)
    return diff


def corresponding_lists(x, y):
    """Returns True is both lists x and y have one to one mapping
        e.g:
            x = [12, 4, 12, 3]
            y = [44, 6, 44, 9]
            corresponding_lists(x, y) # True
    """
    unique_mapping_pairs = set(zip(x, y))
    unique_elements = set(y)
    return len(unique_mapping_pairs) == len(unique_elements)



from nxviz.io import graph_from_dataframe
import pandas as pd

from random import choice, seed

seed(123)
n_nodes = 100

headers = ["a_num", "a_string", "ten_things", "color", "group", "location"]
data_generator = (
    (
        i,
        str(i),
        i % 10,
        choice(["red", "green", "blue"]),
        choice(["A", "B"]),
        choice(
            ["New York", "San Francisco", "Chicago", "LA", "Austin", "Boston"]
        ),
    )
    for i in range(0, n_nodes)
)

test_dataframe = pd.DataFrame(data_generator, columns=headers)


def test_basic_import():

    g = graph_from_dataframe(test_dataframe)

    assert g.number_of_nodes() == 21
    assert g.number_of_edges() == 129
    assert g.nodes["A"] == {"type": "group"}
    assert g["blue"]["New York"] == {"type": "color_location", "weight": 6.0}


def test_percent_threshold():

    g = graph_from_dataframe(test_dataframe, threshold_by_percent_unique=0.08)

    assert g.number_of_nodes() == 11
    assert g.number_of_edges() == 36
    assert g.nodes["A"] == {"type": "group"}
    assert g["blue"]["B"] == {"type": "color_group", "weight": 20.0}


def test_count_threshold():

    g = graph_from_dataframe(test_dataframe, threshold_by_count_unique=3)

    assert g.number_of_nodes() == 5
    assert g.number_of_edges() == 6
    assert g.nodes["A"] == {"type": "group"}
    assert g["blue"]["B"] == {"type": "color_group", "weight": 20.0}


def test_node_properties():
    g = graph_from_dataframe(
        test_dataframe,
        threshold_by_count_unique=3,
        node_property_columns=["ten_things"],
    )

    assert g.number_of_nodes() == 5
    assert g.number_of_edges() == 6
    assert g.nodes["A"] == {
        "ten_things": "4|6|7|8|0|1|2|3|5|9",
        "type": "group",
    }


def test_edge_properties():
    g = graph_from_dataframe(
        test_dataframe, edge_property_columns=["ten_things", "a_string"]
    )

    assert g.number_of_nodes() == 21
    assert g.number_of_edges() == 129
    assert g["blue"]["B"] == {
        "a_string": "96",
        "ten_things": 6,
        "type": "color_group",
        "weight": 20.0,
    }



from setuptools import setup

with open("README.md") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = []
with open("requirements.txt") as rqmts:
    for r in rqmts:
        requirements.append(r.strip("\n"))

setup(
    name="nxviz",
    version="0.6.3",
    description="Graph Visualization Package",
    long_description=readme + "\n\n" + history,
    author="Eric J. Ma",
    author_email="ericmajinglong@gmail.com",
    url="https://github.com/ericmjl/nxviz",
    packages=["nxviz"],
    package_dir={"nxviz": "nxviz"},
    include_package_data=True,
    install_requires=requirements,
    license="MIT license",
    keywords="nxviz",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
)
